import copy
from typing import Any
from typing import Dict


def update_route_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload["properties"]
    if (new_values.get("address_prefix") is not None) and (
        new_values["address_prefix"] != existing_properties.get("addressPrefix")
    ):
        new_payload["properties"]["addressPrefix"] = new_values["address_prefix"]
        need_update = True
    if (new_values.get("next_hop_type") is not None) and (
        new_values["next_hop_type"] != existing_properties.get("nextHopType")
    ):
        new_payload["properties"]["nextHopType"] = new_values["next_hop_type"]
        need_update = True
    if new_values.get("next_hop_ip_address") != existing_properties.get(
        "nextHopIpAddress"
    ):
        new_payload["properties"]["nextHopIpAddress"] = new_values.get(
            "next_hop_ip_address"
        )
        need_update = True
    if need_update:
        result["ret"] = new_payload
    return result


def convert_raw_routes_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    route_table_name: str,
    route_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        route_table_name: Azure Route Table resource name.
        route_name: Azure Route resource name.
        resource_id: Azure Route resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "route_name": route_name,
        "route_table_name": route_table_name,
        "subscription_id": subscription_id,
    }
    properties = resource.get("properties")
    if properties:
        resource_translated["address_prefix"] = properties.get("addressPrefix")
        resource_translated["next_hop_type"] = properties.get("nextHopType")
        if properties.get("nextHopIpAddress") is not None:
            resource_translated["next_hop_ip_address"] = properties.get(
                "nextHopIpAddress"
            )

    return resource_translated


def convert_present_to_raw_routes(
    hub,
    address_prefix: str,
    next_hop_type: str,
    next_hop_ip_address: str = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        address_prefix(str) : The destination CIDR to which the route applies.
        next_hop_type(str) : The type of Azure hop the packet should be sent to.
        next_hop_ip_address(str, optional) : The IP address packets should be forwarded to. Next hop values are only allowed in routes where the next hop type is VirtualAppliance.
    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "properties": {"addressPrefix": address_prefix, "nextHopType": next_hop_type}
    }
    if next_hop_ip_address is not None:
        payload["properties"]["nextHopIpAddress"] = next_hop_ip_address
    return payload



import os
from typing import List

try:
    import bs4  # noqa
    import requests  # noqa
    import tqdm  # noqa

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


__func_alias__ = {"type_": "type"}


def parse(hub, api_url: str, rest_api: str, resources: List[str] = None):
    """
    rest_api can be one of: Azure, Kaizala, Power BI, Playfab, Azure DevOps, Operations Manager, Azure Blockchain Workbench, or Dynamic 365 Online Management
    """
    plugins = {}
    ref = rest_api

    # TODO using selenium should be the default
    # selenium is needed because the html we need is rendered after a bunch of javascript
    USING_SELENIUM = False
    if USING_SELENIUM:
        url = f"https://docs.microsoft.com/en-us/rest/api/{ref.replace(' ', '-')}"
        soup = hub.pop_create.azure.api_reference.parse_url(url, resources)
    else:
        # TODO once selenium is integrated we can remove this and the baked-in file
        if ref.lower() == "azure":
            file_path = os.path.dirname(__file__) + "/resource/azure-rest-api.html"
        else:
            hub.log.error(f"{ref} is not a supported resource.")
        soup = hub.pop_create.azure.api_reference.parse_file(file_path, resources)

    new_plugins = hub.pop_create.azure.api_reference.parse(soup, resources)

    for k, v in new_plugins.items():
        ref = f"{hub.tool.format.case.snake(ref)}.{hub.tool.format.case.snake(k)}"
        plugins[ref] = v

    return plugins



"""State module for managing network Security Rules."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_security_group_name: str,
    security_rule_name: str,
    priority: int,
    direction: str,
    access: str,
    protocol: str,
    description: str = None,
    source_port_range: str = None,
    source_port_ranges: List[str] = None,
    source_address_prefix: str = None,
    source_address_prefixes: List[str] = None,
    source_application_security_groups: List[str] = None,
    destination_port_range: str = None,
    destination_port_ranges: List[str] = None,
    destination_address_prefix: str = None,
    destination_address_prefixes: List[str] = None,
    destination_application_security_groups: List[str] = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Create or update Security Rules.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_security_group_name(str): The name of the network security group.
        security_rule_name(str): The name of the security rule.
        priority(int): The priority of the security rule. The value can be between 100 and 4096.The priority number must
            be unique for each rule in the collection. The lower the priority number, the higher the priority of the rule.
        direction(str): The direction of the rule. The direction specifies if rule will be evaluated on incoming or outgoing traffic.
        access(str): The network traffic is allowed or denied.
        protocol(str): Network protocol this rule applies to.
        description(str, Optional): A description for this rule.
        source_port_range(str, Optional): The source port or range. Integer or range between 0 and 65535. Asterisk '*'
            can also be used to match all ports.
        source_port_ranges(list[str], Optional): The source port ranges. Either this or source_port_range need to be provided.
        source_address_prefix(str, Optional): The source address prefix. CIDR or source IP range. Asterisk '*' can also be used to match all
            source IPs. Default tags such as 'VirtualNetwork', 'AzureLoadBalancer' and 'Internet' can also be used. If
            this is an ingress rule, specifies where network traffic originates from.
        source_address_prefixes(list[str], Optional): The CIDR or source IP ranges.
        source_application_security_groups(list[str], Optional): The list of resource id of the application security group.
            Either one of source_address_prefix, source_address_prefixes or source_application_security_groups needs to be provided.
        destination_port_range(str, Optional): The destination port or range. Integer or range between 0 and 65535.
            Asterisk '*' can also be used to match all ports.
        destination_port_ranges(list[str], Optional): The destination port ranges. Either this or destination_port_range need to be provided.
        destination_address_prefix(str, Optional): The destination address prefix. CIDR or destination IP range.
            Asterisk '*' can also be used to match all source IPs. Default tags such as 'VirtualNetwork',
            'AzureLoadBalancer' and 'Internet' can also be used.
        destination_address_prefixes(list[str], Optional): The destination address prefixes. CIDR or destination IP ranges.
        destination_application_security_groups(list[str], Optional): The list of resource id of the application security group.
            Either one of destination_address_prefix, destination_address_prefixes or destination_application_security_groups needs to be provided.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Security rule resource id.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.network.security_rules.present:
                - name: rule1
                - resource_id: /subscriptions/12345678-1234-1234-1234-aaabc1234aaa/resourceGroups/rg-1/providers/Microsoft.Network/networkSecurityGroups/nsg-1/securityRules/rule1
                - resource_group_name: rg-1
                - subscription_id: 12345678-1234-1234-1234-aaabc1234aaa
                - network_security_group_name: nsg-1
                - security_rule_name: rule1
                - protocol: '*'
                - source_address_prefix: '*'
                - destination_port_range: '80'
                - destination_address_prefix: '*'
                - access: Allow
                - priority: 130
                - direction: Inbound
                - source_port_ranges:
                  - '20'
                  - '50'
                  - '80'
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id

    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}/securityRules/{security_rule_name}"

    response_get = await hub.exec.azure.network.security_rules.get(
        ctx, resource_id=resource_id, raw=True
    )

    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.network.security_rules {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result

    if not response_get["ret"]:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_group_name": resource_group_name,
                    "network_security_group_name": network_security_group_name,
                    "security_rule_name": security_rule_name,
                    "priority": priority,
                    "direction": direction,
                    "access": access,
                    "protocol": protocol,
                    "description": description,
                    "source_port_range": source_port_range,
                    "source_port_ranges": source_port_ranges,
                    "source_address_prefix": source_address_prefix,
                    "source_address_prefixes": source_address_prefixes,
                    "source_application_security_groups": source_application_security_groups,
                    "destination_port_range": destination_port_range,
                    "destination_port_ranges": destination_port_ranges,
                    "destination_address_prefix": destination_address_prefix,
                    "destination_address_prefixes": destination_address_prefixes,
                    "destination_application_security_groups": destination_application_security_groups,
                    "subscription_id": subscription_id,
                    "resource_id": resource_id,
                },
            )
            result["comment"].append(
                f"Would create azure.network.security_rules '{name}'"
            )
            return result

        else:
            # PUT operation to create a resource
            payload = hub.tool.azure.network.security_rules.convert_present_to_raw_security_rules(
                priority=priority,
                direction=direction,
                access=access,
                protocol=protocol,
                description=description,
                source_port_range=source_port_range,
                source_port_ranges=source_port_ranges,
                source_address_prefix=source_address_prefix,
                source_address_prefixes=source_address_prefixes,
                source_application_security_groups=source_application_security_groups,
                destination_port_range=destination_port_range,
                destination_port_ranges=destination_port_ranges,
                destination_address_prefix=destination_address_prefix,
                destination_address_prefixes=destination_address_prefixes,
                destination_application_security_groups=destination_application_security_groups,
            )
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{hub.exec.azure.URL}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}/securityRules/{security_rule_name}?api-version=2022-07-01",
                success_codes=[200, 201],
                json=payload,
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create azure.network.security_rules {response_put['comment']} {response_put['ret']}"
                )
                result["comment"] = [response_put["comment"], response_put["ret"]]
                result["result"] = False
                return result

            result[
                "new_state"
            ] = hub.tool.azure.network.security_rules.convert_raw_security_rules_to_present(
                resource=response_put["ret"], idem_resource_name=name
            )
            result["comment"].append(f"Created azure.network.security_rules '{name}'")
            return result
    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.network.security_rules.convert_raw_security_rules_to_present(
            resource=existing_resource, idem_resource_name=name
        )
        # Generate a new PUT operation payload with new values
        new_payload = hub.tool.azure.network.security_rules.update_security_rules_payload(
            response_get["ret"],
            {
                "priority": priority,
                "direction": direction,
                "access": access,
                "protocol": protocol,
                "description": description,
                "source_port_range": source_port_range,
                "source_port_ranges": source_port_ranges,
                "source_address_prefix": source_address_prefix,
                "source_address_prefixes": source_address_prefixes,
                "source_application_security_groups": source_application_security_groups,
                "destination_port_range": destination_port_range,
                "destination_port_ranges": destination_port_ranges,
                "destination_address_prefix": destination_address_prefix,
                "destination_address_prefixes": destination_address_prefixes,
                "destination_application_security_groups": destination_application_security_groups,
            },
        )
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.network.security_rules '{name}' has no property need to be updated."
            )
            return result
        else:
            if ctx.get("test", False):
                result[
                    "new_state"
                ] = hub.tool.azure.network.security_rules.convert_raw_security_rules_to_present(
                    resource=new_payload["ret"], idem_resource_name=name
                )
                result["comment"].append(
                    f"Would update azure.network.security_rules '{name}'"
                )
                return result
            else:
                # PUT call to update the resource
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{hub.exec.azure.URL}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}/securityRules/{security_rule_name}?api-version=2022-07-01",
                    success_codes=[200, 201],
                    json=new_payload["ret"],
                )

                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not update azure.network.security_rules {response_put['comment']} {response_put['ret']}"
                    )
                    result["result"] = False
                    result["comment"] = [
                        response_put["comment"],
                        response_put["ret"],
                    ]
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.security_rules.convert_raw_security_rules_to_present(
                    resource=response_put["ret"], idem_resource_name=name
                )
                result["comment"].append(
                    f"Updated azure.network.security_rules '{name}'"
                )
                return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_security_group_name: str,
    security_rule_name: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    r"""Delete Security Rule.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_security_group_name(str): The name of the network security group.
        security_rule_name(str): The name of the security rule.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.security_rules.absent:
                - name: value
                - resource_group_name: value
                - network_security_group_name: value
                - security_rule_name: value
    """
    result = dict(name=name, result=True, comment=[], old_state=None, new_state=None)
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}/securityRules/{security_rule_name}"

    response_get = await hub.exec.azure.network.security_rules.get(
        ctx,
        resource_id=resource_id,
    )
    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.network.security_rules {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result

    if response_get["ret"]:
        result["old_state"] = response_get["ret"]
        result["old_state"]["name"] = name
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.network.security_rules '{name}'"
            )
            return result

        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{hub.exec.azure.URL}{resource_id}?api-version=2022-07-01",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.network.security_rules {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.network.security_rules '{name}'")
        return result

    else:
        # If Azure returns 'Not Found' error, it means the resource is absent.
        result["comment"].append(
            f"azure.network.security_rules '{name}' already absent"
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Security Rules under the same subscription.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.security_rules
    """
    result = {}
    ret_list = await hub.exec.azure.network.security_rules.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(
            f"Could not describe network security_rules {ret_list['comment']}"
        )
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.security_rules.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_security_rules_to_present(
    hub, resource: Dict, idem_resource_name: str
) -> Dict[str, Any]:
    """
    Converts raw security_rules api response into present function format

    Args:
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_id = resource["id"]
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "networkSecurityGroups": "network_security_group_name",
            "securityRules": "security_rule_name",
        }
    )
    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )
    translated_resource = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        **uri_parameter_values,
    }
    properties = resource.get("properties")
    properties_parameters = {
        "protocol": "protocol",
        "description": "description",
        "sourcePortRange": "source_port_range",
        "sourceAddressPrefix": "source_address_prefix",
        "destinationPortRange": "destination_port_range",
        "destinationAddressPrefix": "destination_address_prefix",
        "access": "access",
        "priority": "priority",
        "direction": "direction",
        "sourcePortRanges": "source_port_ranges",
        "destinationPortRanges": "destination_port_ranges",
        "sourceAddressPrefixes": "source_address_prefixes",
        "destinationAddressPrefixes": "destination_address_prefixes",
        "sourceApplicationSecurityGroups": "source_application_security_groups",
        "destinationApplicationSecurityGroups": "destination_application_security_groups",
    }
    for parameter_raw, parameter_present in properties_parameters.items():
        if properties.get(parameter_raw):
            if (
                parameter_present == "destination_application_security_groups"
                or parameter_present == "source_application_security_groups"
            ):
                translated_resource[parameter_present] = [
                    value["id"] for value in properties.get(parameter_raw)
                ]
            else:
                translated_resource[parameter_present] = properties.get(parameter_raw)
    return translated_resource


def convert_present_to_raw_security_rules(
    hub,
    priority: int,
    direction: str,
    access: str,
    protocol: str,
    description: str,
    source_port_range: str = None,
    source_port_ranges: List = None,
    source_address_prefix: str = None,
    source_address_prefixes: List = None,
    source_application_security_groups: List = None,
    destination_port_range: str = None,
    destination_port_ranges: List = None,
    destination_address_prefix: str = None,
    destination_address_prefixes: List = None,
    destination_application_security_groups: List = None,
) -> Dict[str, Any]:
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        priority(int): The priority of the security rule. The value can be between 100 and 4096.The priority number must
            be unique for each rule in the collection. The lower the priority number, the higher the priority of the rule.
        direction(str): The direction of the rule. The direction specifies if rule will be evaluated on incoming or outgoing traffic.
        access(str): The network traffic is allowed or denied.
        protocol(str): Network protocol this rule applies to.
        description(str, Optional): A description for this rule.
        source_port_range(str, Optional): The source port or range. Integer or range between 0 and 65535. Asterisk '*'
            can also be used to match all ports.
        source_port_ranges(List, Optional): The source port ranges. Either this or source_port_range need to be provided.
        source_address_prefix(str, Optional): The source address prefix. CIDR or source IP range. Asterisk '*' can also be used to match all
            source IPs. Default tags such as 'VirtualNetwork', 'AzureLoadBalancer' and 'Internet' can also be used. If
            this is an ingress rule, specifies where network traffic originates from.
        source_address_prefixes(List, Optional): The CIDR or source IP ranges.
        source_application_security_groups(List, Optional): The list of resource id of the application security group.
            Either one of source_address_prefix, source_address_prefixes or source_application_security_groups needs to be provided.
        destination_port_range(str, Optional): The destination port or range. Integer or range between 0 and 65535.
            Asterisk '*' can also be used to match all ports.
        destination_port_ranges(List, Optional): The destination port ranges. Either this or destination_port_range need to be provided.
        destination_address_prefix(str, Optional): The destination address prefix. CIDR or destination IP range.
            Asterisk '*' can also be used to match all source IPs. Default tags such as 'VirtualNetwork',
            'AzureLoadBalancer' and 'Internet' can also be used.
        destination_address_prefixes(List, Optional): The destination address prefixes. CIDR or destination IP ranges.
        destination_application_security_groups(List, Optional): The list of resource id of the application security group.
            Either one of destination_address_prefix, destination_address_prefixes or destination_application_security_groups needs to be provided.



    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {"properties": {}}
    properties_parameters = {
        "protocol": "protocol",
        "description": "description",
        "sourcePortRange": "source_port_range",
        "sourceAddressPrefix": "source_address_prefix",
        "destinationPortRange": "destination_port_range",
        "destinationAddressPrefix": "destination_address_prefix",
        "access": "access",
        "priority": "priority",
        "direction": "direction",
        "sourcePortRanges": "source_port_ranges",
        "destinationPortRanges": "destination_port_ranges",
        "sourceAddressPrefixes": "source_address_prefixes",
        "destinationAddressPrefixes": "destination_address_prefixes",
        "sourceApplicationSecurityGroups": "source_application_security_groups",
        "destinationApplicationSecurityGroups": "destination_application_security_groups",
    }
    for parameter_raw, parameter_present in properties_parameters.items():
        if locals()[parameter_present]:
            if (
                parameter_present == "destination_application_security_groups"
                or parameter_present == "source_application_security_groups"
            ):
                payload["properties"][parameter_raw] = []
                for value in locals()[parameter_present]:
                    payload["properties"][parameter_raw].append({"id": value})
            else:
                payload["properties"][parameter_raw] = locals()[parameter_present]
    return payload


def update_security_rules_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload["properties"]
    properties_parameters = {
        "protocol": "protocol",
        "description": "description",
        "sourcePortRange": "source_port_range",
        "sourceAddressPrefix": "source_address_prefix",
        "destinationPortRange": "destination_port_range",
        "destinationAddressPrefix": "destination_address_prefix",
        "access": "access",
        "priority": "priority",
        "direction": "direction",
        "sourcePortRanges": "source_port_ranges",
        "destinationPortRanges": "destination_port_ranges",
        "sourceAddressPrefixes": "source_address_prefixes",
        "destinationAddressPrefixes": "destination_address_prefixes",
        "sourceApplicationSecurityGroups": "source_application_security_groups",
        "destinationApplicationSecurityGroups": "destination_application_security_groups",
    }
    optional_params = [
        "sourcePortRange",
        "sourceAddressPrefix",
        "destinationPortRange",
        "destinationAddressPrefix",
        "sourcePortRanges",
        "destinationPortRanges",
        "sourceAddressPrefixes",
        "destinationAddressPrefixes",
        "sourceApplicationSecurityGroups",
        "destinationApplicationSecurityGroups",
    ]

    for param in optional_params:
        new_payload["properties"].pop(param, None)

    for parameter_raw, parameter_present in properties_parameters.items():
        if new_values.get(parameter_present):
            if parameter_present in [
                "destination_application_security_groups",
                "source_application_security_groups",
            ]:
                if existing_properties.get(parameter_raw) and (
                    sorted(new_values.get(parameter_present))
                    == sorted(
                        [
                            value["id"]
                            for value in existing_properties.get(parameter_raw)
                        ]
                    )
                ):
                    new_payload["properties"][parameter_raw] = existing_properties.get(
                        parameter_raw
                    )
                else:
                    new_payload["properties"][parameter_raw] = []
                    for value in new_values.get(parameter_present):
                        new_payload["properties"][parameter_raw].append({"id": value})
                    need_update = True
            elif isinstance(new_values.get(parameter_present), List):
                if existing_properties.get(parameter_raw) and (
                    sorted(new_values.get(parameter_present))
                    == sorted(existing_properties.get(parameter_raw))
                ):
                    new_payload["properties"][parameter_raw] = existing_properties.get(
                        parameter_raw
                    )
                else:
                    new_payload["properties"][parameter_raw] = new_values.get(
                        parameter_present
                    )
                    need_update = True
            else:
                if new_values.get(parameter_present) != existing_properties.get(
                    parameter_raw
                ):
                    new_payload["properties"][parameter_raw] = new_values.get(
                        parameter_present
                    )
                    need_update = True
                else:
                    new_payload["properties"][parameter_raw] = existing_properties.get(
                        parameter_raw
                    )

    if need_update:
        result["ret"] = new_payload
    return result



"""Exec module for managing Attach Subscriptions"""

__func_alias__ = {"list_": "list"}

from typing import Dict
from collections import OrderedDict


async def get(hub, ctx, resource_id: str, raw: bool = False) -> Dict:
    """Gets the subscription attached to a management group from azure account.

    Args:
        resource_id(str):
            The resource id of the subscription attached to a management group.
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.subscription.attach_subscriptions.get resource_id="/providers/Microsoft.Management/managementGroups/{management_group_id}/subscriptions/{subscription_id}" raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.subscription.attach_subscriptions.get
                - kwargs:
                    resource_id: /providers/Microsoft.Management/managementGroups/{management_group_id}/subscriptions/{subscription_id}
                    raw: False
    """

    result = dict(comment=[], ret=None, result=True)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"] = response_get["comment"]
        return result
    uri_parameters = OrderedDict({"managementGroups": "management_group_id"})
    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        response_get["ret"]["properties"]["parent"]["id"], uri_parameters
    )
    if raw:
        result["ret"] = response_get["ret"]
    else:
        result[
            "ret"
        ] = hub.tool.azure.subscription.subscriptions.convert_raw_attach_subscription_to_present(
            resource=response_get["ret"],
            idem_resource_name=resource_id,
            resource_id=resource_id,
            subscription_id=response_get["ret"]["name"],
            **uri_parameter_values,
        )

    return result


async def list_(
    hub,
    ctx,
) -> Dict:
    """Lists all Attached Subscriptions.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.subscription.attach_subscriptions.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.subscription.attach_subscriptions.list

    """

    result = dict(comment=[], ret=[], result=True)

    uri_parameters = OrderedDict({"managementGroups": "management_group_id"})
    ret = await hub.exec.request.json.post(
        ctx,
        url=f"{ctx.acct.endpoint_url}/providers/Microsoft.Management/getEntities?api-version=2020-05-01&$view=SubscriptionsOnly",
        success_codes=[200],
    )
    if not ret["result"]:
        hub.log.debug("Could not describe attach_subscription")
    resource_list = ret["ret"].get("value", None)
    if resource_list:
        for resource in resource_list:
            subscription_id = resource["name"]
            uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                resource["properties"]["parent"]["id"], uri_parameters
            )
            resource_id = f"/providers/MicrosoftManagement/managementGroups/{uri_parameter_values.get('management_group_id')}/subscriptions/{subscription_id}"
            result["ret"].append(
                hub.tool.azure.subscription.subscriptions.convert_raw_attach_subscription_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                    **uri_parameter_values,
                )
            )
    return result



"""State module for managing Route."""
import copy
from typing import Any
from typing import Dict


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    route_table_name: str,
    route_name: str,
    address_prefix: str,
    next_hop_type: str,
    next_hop_ip_address: str = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> dict:
    r"""Create or update Routes.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        route_table_name(str): The name of the route table.
        route_name(str): The name of the route.
        address_prefix(str): Resource location. This field can not be updated.
        next_hop_type(str): The type of Azure hop the packet should be sent to.
        next_hop_ip_address(str, Optional): The IP address packets should be forwarded to. Next hop values are only allowed in routes where the next hop type is VirtualAppliance.
        subscription_id: Subscription Unique id.
        resource_id: Route table resource id on Azure.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.network.routes.present:
              - name: resource_name
              - resource_group_name: my-resource-group
              - route_name: my-route
              - route_table_name: my-route-table
              - address_prefix:10.0.0.0/25
              - next_hop_ip_address: 10.0.1.0
              - next_hop_type: VirtualAppliance
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/routeTables/{route_table_name}/routes/{route_name}"
    response_get = await hub.exec.azure.network.routes.get(
        ctx,
        resource_id=resource_id,
        raw=True,
    )
    if response_get["result"]:
        if response_get["ret"] is None:
            if ctx.get("test", False):
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "route_table_name": route_table_name,
                        "route_name": route_name,
                        "address_prefix": address_prefix,
                        "next_hop_type": next_hop_type,
                        "next_hop_ip_address": next_hop_ip_address,
                        "resource_id": resource_id,
                        "subscription_id": subscription_id,
                    },
                )
                result["comment"].append(f"Would create azure.network.routes '{name}'")
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.network.routes.convert_present_to_raw_routes(
                    address_prefix=address_prefix,
                    next_hop_type=next_hop_type,
                    next_hop_ip_address=next_hop_ip_address,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.network.routes {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result
                result[
                    "new_state"
                ] = hub.tool.azure.network.routes.convert_raw_routes_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    route_table_name=route_table_name,
                    route_name=route_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(f"Created azure.network.routes '{name}'")
                return result
        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.network.routes.convert_raw_routes_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                route_name=route_name,
                route_table_name=route_table_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.network.routes.update_route_payload(
                existing_resource,
                {
                    "address_prefix": address_prefix,
                    "next_hop_type": next_hop_type,
                    "next_hop_ip_address": next_hop_ip_address,
                },
            )
            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.network.routes '{name}' has no property need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.network.routes.convert_raw_routes_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_group_name=resource_group_name,
                        route_table_name=route_table_name,
                        route_name=route_name,
                        resource_id=resource_id,
                        subscription_id=subscription_id,
                    )

                    result["comment"].append(
                        f"Would update azure.network.routes '{name}'"
                    )
                return result
            # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.network.routes '{name}' has no property need to be updated."
                )
                return result
            result["comment"] += new_payload["comment"]
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                success_codes=[200],
                json=new_payload["ret"],
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.network.routes {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_put["comment"], response_put["ret"]]
                return result

            result[
                "new_state"
            ] = hub.tool.azure.network.routes.convert_raw_routes_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                route_table_name=route_table_name,
                route_name=route_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            result["comment"].append(f"Updated azure.network.routes '{name}'")
            return result
    else:
        hub.log.debug(
            f"Could not get azure.network.routes {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


__contracts__ = ["resource"]


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    route_table_name: str,
    route_name: str,
    subscription_id: str = None,
) -> dict:
    r"""Delete Routes.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        route_name(str): The name of the route.
        route_table_name(str): The name of the route table.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.routes.absent:
                - name: my-resource
                - resource_group_name: my-resource-group-name
                - route_name: my-route
                - route_table_name: my-route-table
    """
    result = dict(name=name, result=True, comment=[], old_state=None, new_state=None)
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/routeTables/{route_table_name}/routes/{route_name}"
    response_get = await hub.exec.azure.network.routes.get(
        ctx,
        resource_id=resource_id,
        raw=True,
    )
    if response_get["result"]:
        if response_get["ret"]:
            result[
                "old_state"
            ] = hub.tool.azure.network.routes.convert_raw_routes_to_present(
                resource=response_get["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                route_name=route_name,
                route_table_name=route_table_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            if ctx.get("test", False):
                result["comment"].append(f"Would delete azure.network.routes '{name}'")
                return result
            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                success_codes=[200, 202, 204],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"Could not delete azure.network.routes {response_delete['comment']} {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_delete["comment"], response_delete["ret"]]
                return result

            result["comment"].append(f"Deleted azure.network.routes '{name}'")
            return result
        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(f"azure.network.routes '{name}' already absent")
            return result
    else:
        hub.log.debug(
            f"Could not get azure.network.routes {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Route under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.routes
    """
    result = {}
    ret = await hub.exec.azure.network.routes.list(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe network routes {ret['comment']}")
        return {}

    for route in ret["ret"]:
        resource_id = route.get("resource_id")
        result[resource_id] = {
            "azure.network.routes.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in route.items()
            ]
        }
    return result



try:
    import validators

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


# https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
SUPPORTED_HTTP_METHODS = {
    "GET",
    "HEAD",
    "POST",
    "PUT",
    "DELETE",
    "CONNECT",
    "OPTIONS",
    "TRACE",
    "PATCH",
}


def request_method(hub, http_method: str):
    if http_method is None or http_method.upper() not in SUPPORTED_HTTP_METHODS:
        raise ValueError(f"{http_method} is not a valid http request method.")


def url(hub, http_url: str):
    try:
        validators.url(http_url)
    except validators.ValidationFailure as e:
        raise ValueError(f"{http_url} is not a valid http url")



import copy
from typing import Any
from typing import Dict
from typing import List


def update_route_table_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload["properties"]
    if (new_values.get("disable_bgp_route_propagation") is not None) and (
        new_values["disable_bgp_route_propagation"]
        != existing_properties.get("disableBgpRoutePropagation")
    ):
        new_payload["properties"]["disableBgpRoutePropagation"] = new_values[
            "disable_bgp_route_propagation"
        ]
        need_update = True
    if (new_values.get("tags") is not None) and (
        new_values["tags"] != existing_payload.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        need_update = True
    if new_values["routes"] is not None:
        if len(existing_properties.get("routes")) == 0 or compare_routes(
            new_values["routes"],
            convert_raw_routes_to_present(existing_properties.get("routes")),
            "route_name",
        ):
            new_payload["properties"]["routes"] = convert_present_to_raw_routes(
                new_values["routes"]
            )
            need_update = True
    if need_update:
        result["ret"] = new_payload
    return result


def convert_raw_route_tables_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    route_table_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        route_table_name: Azure Route Table resource name.
        resource_id: Azure Subnet resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "route_table_name": route_table_name,
        "location": resource["location"],
        "subscription_id": subscription_id,
    }
    if resource.get("tags") is not None:
        resource_translated["tags"] = resource["tags"]
    properties = resource.get("properties")
    if properties:
        if properties.get("disableBgpRoutePropagation") is not None:
            resource_translated["disable_bgp_route_propagation"] = properties.get(
                "disableBgpRoutePropagation"
            )
        if len(properties.get("routes")) != 0:
            resource_translated["routes"] = convert_raw_routes_to_present(
                properties.get("routes")
            )
    return resource_translated


def convert_present_to_raw_route_tables(
    hub,
    location: str,
    disable_bgp_route_propagation: bool = None,
    routes: List = None,
    tags: str = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location: Resource location. This field can not be updated.
        disable_bgp_route_propagation:  Whether to disable the routes learned by BGP on that route table. True means disable.
        routes: An array of routes on the route table. Collection of routes contained within a route table.
         Defaults to None
         * route_name(str) -- The name of the resource that is unique within a resource group.
         * address_prefix(str) -- The destination CIDR to which the route applies.
         * next_hop_type(str) -- The type of Azure hop the packet should be sent to.
         * next_hop_ip_address(str, optional) -- The IP address packets should be forwarded to. Next hop values are only allowed in routes where the next hop type is VirtualAppliance.
        tags: Resource tags.
    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location, "properties": dict()}
    if tags is not None:
        payload["tags"] = tags
    if disable_bgp_route_propagation is not None:
        payload["properties"][
            "disableBgpRoutePropagation"
        ] = disable_bgp_route_propagation
    if routes is not None:
        payload["properties"]["routes"] = convert_present_to_raw_routes(routes)
    return payload


def convert_present_to_raw_routes(routes: List):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        routes(list[dict], required): List of Route in a route table resource.Each Route will have fields
        ((route_name(str), required), (address_prefix(str), required), (next_hop_type(str), required) and (next_hop_ip_address(str), optional))

    Returns:
        Routes List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    raw_routes: List = []
    for route in routes:
        new_route_payload = {
            "name": route["route_name"],
            "properties": {
                "addressPrefix": route["address_prefix"],
                "nextHopType": route["next_hop_type"],
            },
        }
        if route.get("next_hop_ip_address") is not None:
            new_route_payload["properties"]["nextHopIpAddress"] = route[
                "next_hop_ip_address"
            ]
        raw_routes.append(new_route_payload)
    return raw_routes


def convert_raw_routes_to_present(routes: List):
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
    present input parameters.

    Args:
        routes(list[dict], optional): Resource List of Routes in a Route Table resource.

    Returns:
         A route List that contains the parameters that match respective present function's input format.
    """
    present_routes: List = []
    for route in routes:
        new_route_payload = {
            "route_name": route["name"],
            "address_prefix": route["properties"]["addressPrefix"],
            "next_hop_type": route["properties"]["nextHopType"],
        }
        if route.get("properties").get("nextHopIpAddress") is not None:
            new_route_payload["next_hop_ip_address"] = route["properties"][
                "nextHopIpAddress"
            ]
        present_routes.append(new_route_payload)
    return present_routes


def compare_routes(
    new_values: List[Dict[str, any]], resource: List[Dict[str, any]], sorting_key: str
):
    """
    Compares routes to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        new_values: (List[Dict[str, any]]) Present value which will be given as input
        resource: (List[Dict[str, any]]) Raw resource response which needs to be compared with new_values
        sorting_key: (str) Primary/Unique key name within each dictionary , which will be used to sort dictionary
         objects with given list before comparing.

    Returns:
        A boolean value, True if there is any difference between List[Dict] arguments else returns False
    """
    if len(new_values) == 0 or len(resource) == 0:
        return True
    return sorted(new_values, key=lambda x: x[sorting_key]) != sorted(
        resource, key=lambda x: x[sorting_key]
    )



"""State module for managing Subscription."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 20}}


async def present(
    hub,
    ctx,
    name: str,
    alias: str,
    billing_scope: str,
    display_name: str,
    workload: str,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Subscription.

    Args:
        name(str): The identifier for this state.
        alias(str): The alias name of the subscription to create or update. Can include alphanumeric,
          underscore, parentheses, hyphen, period (except at end), and Unicode characters that match
          the allowed characters.Regex pattern: ^[-\w\._\(\)]+$.
        billing_scope(str): billing scope associated with billing account id and enrollment account id to create subscription
        display_name(str): subscription display name
        workload(str): type of workload where subscription needs to be created. Can be Production/DevTest
        resource_id(str, Optional): resource unique id

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_present:
              azure.subscription.subscriptions.present:
                - name: create_subscription_3
                - alias: dupSubTest
                - billing_scope: /providers/Microsoft.Billing/billingAccounts/{billingAccountID}/enrollmentAccounts/{enrollmentAccountId}}
                - display_name: Subscription Display Name
                - workload: Production
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if resource_id is None:
        resource_id = f"/providers/Microsoft.Subscription/aliases/{alias}"
    response_get = await hub.exec.azure.subscription.subscriptions.get(
        ctx, resource_id=resource_id, raw=True
    )
    hub.log.debug(f"Get Subscription details : {response_get}")

    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.subscriptions.subscription '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result

    if not response_get["ret"]:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "alias": alias,
                    "display_name": display_name,
                    "subscription_id": None,
                    "resource_id": resource_id,
                },
            )
            result["comment"].append(
                f"Would create azure.subscription.subscriptions '{name}'"
            )
            return result
        else:
            # PUT operation to create subscription
            payload = hub.tool.azure.subscription.subscriptions.convert_present_to_raw_subscription(
                billing_scope=billing_scope,
                display_name=display_name,
                workload=workload,
            )
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-09-01",
                success_codes=[200, 201],
                json=payload,
            )
            if not response_put["result"]:
                result["comment"] = [response_put["comment"], response_put["ret"]]
                result["result"] = False
                return result

        result[
            "new_state"
        ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            alias=alias,
            display_name=display_name,
            subscription_id=None,
            resource_id=resource_id,
        )
        result["comment"].append(f"Created azure.subscription.subscriptions '{name}'")
        return result

    else:
        # Update/rename subscription display name if given subscription already present
        # and display name is modified in the input
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            alias=alias,
            display_name=None,
            subscription_id=None,
            resource_id=resource_id,
        )
        subscription_id = result["old_state"]["subscription_id"]
        # Generate a new PUT operation payload with new values
        is_display_name_updated = display_name != existing_resource.get("displayName")
        if ctx.get("test", False):
            if not is_display_name_updated:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.subscription.subscriptions '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
                    resource=response_get["ret"],
                    idem_resource_name=name,
                    alias=alias,
                    display_name=display_name,
                    subscription_id=None,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Would update azure.subscription.subscriptions '{name}'"
                )
            return result
        # POST operation to rename subscription display_name
        if not is_display_name_updated:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.subscription.subscriptions '{name}' has no property need to be updated."
            )
            return result
        rename_payload = {"subscriptionName": display_name}
        response_rename = await hub.exec.request.json.post(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Subscription/rename?api-version=2020-09-01",
            success_codes=[200, 201],
            json=rename_payload,
        )
        if not response_rename["result"]:
            hub.log.debug(
                f"Could not update azure.subscription.subscriptions {response_rename['comment']} {response_rename['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_rename["comment"], response_rename["ret"]]
            return result

        result["new_state"] = copy.deepcopy(result["old_state"])
        result["new_state"]["display_name"] = display_name
        result["comment"].append(f"Updated azure.subscription.subscriptions '{name}'")
        return result


async def absent(hub, ctx, name: str, alias: str) -> Dict:
    r"""Delete Subscription. This state disables the subscription and deletes respective alias.

    Once subscription is in disabled state, after 90 days it gets deleted automatically.

    Args:
        name(str): The identifier for this state.
        alias(str): The alias name of the subscription to delete.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            subscription_is_absent:
              azure.subscription.subscriptions.absent:
                - name: value
                - alias: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    resource_id = f"/providers/Microsoft.Subscription/aliases/{alias}"
    response_get = await hub.exec.azure.subscription.subscriptions.get(
        ctx,
        resource_id=resource_id,
    )
    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.subscriptions.subscription '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result

    if response_get["ret"]:
        result["old_state"] = response_get["ret"]
        result["old_state"]["name"] = name
        subscription_id = response_get["ret"]["subscription_id"]
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.subscription.subscriptions '{name}'"
            )
            return result

        # Cancel the subscription before deleting
        # Once subscription is cancelled, it will be in disabled state and after 90 days it gets deleted automatically.
        response_cancel = await hub.exec.request.json.post(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Subscription/cancel?api-version=2020-09-01",
            success_codes=[200, 202],
        )
        if not response_cancel["result"]:
            result["result"] = False
            result["comment"] = [response_cancel["comment"], response_cancel["ret"]]
            return result

        # Delete subscription alias.
        # After deleting the subscription with alias, deleted alias can be used to create new subscription
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-09-01",
            success_codes=[200, 202],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.subscription.subscriptions {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.subscription.subscriptions '{name}'")
        return result
    else:
        # If Azure returns 'Not Found' error, it means the management group is not associated with given subscription.
        result["comment"].append(
            f"azure.subscriptions.subscription '{name}' already absent"
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe subscriptions in a way that can be recreated/managed with the corresponding "present" function.

    Lists all subscriptions.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            $ idem describe azure.subscription.subscriptions
    """
    result = {}
    ret_list = await hub.exec.azure.subscription.subscriptions.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(
            f"Could not describe subscription subscriptions {ret_list['comment']}"
        )
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.subscription.subscriptions.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


def convert_present_to_raw_virtual_machine(
    hub,
    location: str,
    virtual_machine_size: str,
    network_interface_ids: List[str],
    storage_image_reference: Dict[str, Any],
    storage_os_disk: Dict[str, Any],
    storage_data_disks: List[Dict[str, Any]],
    os_profile: Dict[str, Any],
    tags: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location(str): Resource location. Changing this forces a new resource to be created.
        virtual_machine_size(str): Specifies the size of the Virtual Machine.
        network_interface_ids(List[str]): A list of Network Interface IDs which should be associated with the Virtual Machine.
        tags(Dict, optional): Resource tags.
        os_profile(Dict): Specifies the operating system settings used while creating the virtual machine.
        storage_image_reference(Dict): Specifies information about the image to use. Eg- platform images, marketplace images.
        storage_os_disk(Dict): Specifies information about the operating system disk used by the virtual machine.
        storage_data_disks(list(Dict), optional): List of Data disks attached/added to a VM.

    Returns:
        A Dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "location": location,
        "properties": {"hardwareProfile": {"vmSize": virtual_machine_size}},
    }
    if network_interface_ids is not None:
        network_interface_ids_payload = {
            "networkInterfaces": convert_present_to_raw_network_interfaces(
                network_interface_ids
            )
        }
        payload["properties"]["networkProfile"] = network_interface_ids_payload

    storage_profile_payload = {}
    if storage_image_reference is not None:
        storage_profile_payload[
            "imageReference"
        ] = convert_present_to_raw_image_reference(storage_image_reference)
    if storage_os_disk is not None:
        storage_profile_payload["osDisk"] = convert_present_to_raw_os_disk(
            storage_os_disk
        )
    if storage_data_disks is not None:
        storage_profile_payload["dataDisks"] = convert_present_to_raw_data_disks(
            storage_data_disks
        )
    payload["properties"]["storageProfile"] = storage_profile_payload

    if os_profile is not None:
        payload["properties"]["osProfile"] = convert_present_to_raw_os_profile(
            os_profile
        )

    if tags is not None:
        payload["tags"] = tags

    return payload


def convert_raw_virtual_machine_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    virtual_machine_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        virtual_machine_name: Azure Virtual Machine resource name.
        resource_id: Azure Virtual Machine resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
      A Dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "virtual_machine_name": virtual_machine_name,
        "location": resource.get("location"),
        "subscription_id": subscription_id,
    }
    properties = resource.get("properties")
    if properties:
        if properties.get("hardwareProfile") is not None:
            resource_translated["virtual_machine_size"] = properties.get(
                "hardwareProfile"
            ).get("vmSize")
        if properties.get("osProfile") is not None:
            os_profile_payload = convert_raw_to_present_os_profile(
                os_profile=properties.get("osProfile")
            )
            resource_translated["os_profile"] = os_profile_payload

        if properties.get("networkProfile") is not None:
            if properties.get("networkProfile").get("networkInterfaces") is not None:
                network_profile_payload = convert_raw_to_present_network_interface(
                    network_interfaces=properties.get("networkProfile").get(
                        "networkInterfaces"
                    )
                )
                resource_translated["network_interface_ids"] = network_profile_payload

        if properties.get("storageProfile") is not None:
            storage_profile_properties = properties.get("storageProfile")
            if storage_profile_properties.get("imageReference") is not None:
                image_reference_payload = convert_raw_to_present_image_reference(
                    image_reference=storage_profile_properties.get("imageReference")
                )
                resource_translated["storage_image_reference"] = image_reference_payload
            if storage_profile_properties.get("osDisk") is not None:
                os_disk_payload = convert_raw_to_present_os_disk(
                    os_disk=storage_profile_properties.get("osDisk")
                )
                resource_translated["storage_os_disk"] = os_disk_payload
            if storage_profile_properties.get("dataDisks") is not None:
                data_disk_payload = convert_raw_to_present_data_disks(
                    data_disks=storage_profile_properties.get("dataDisks")
                )
                resource_translated["storage_data_disks"] = data_disk_payload
    if "tags" in resource:
        resource_translated["tags"] = resource.get("tags")
    return resource_translated


def update_virtual_machine_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
    PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
        this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
        do an update.

    Returns:
        A result Dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload.get("properties")

    updated_properties_payload = new_payload.get("properties")
    if new_values.get("virtual_machine_size") is not None:
        if (
            (existing_properties.get("hardwareProfile") is not None)
            and (existing_properties.get("hardwareProfile").get("vmSize") is not None)
            and (
                existing_properties.get("hardwareProfile").get("vmSize")
                != new_values.get("virtual_machine_size")
            )
        ):
            updated_properties_payload["hardwareProfile"]["vmSize"] = new_values.get(
                "virtual_machine_size"
            )
            need_update = True

    if new_values.get("network_interface_ids"):
        if (
            (existing_properties.get("networkProfile") is not None)
            and (
                existing_properties.get("networkProfile").get("networkInterfaces")
                is not None
            )
            and (
                compare_network_interface_payload(
                    existing_properties.get("networkProfile").get("networkInterfaces"),
                    new_values.get("network_interface_ids"),
                )
            )
        ):
            network_profile_payload = {
                "networkInterfaces": convert_present_to_raw_network_interfaces(
                    network_interface_ids=new_values.get("network_interface_ids")
                )
            }
            updated_properties_payload["networkProfile"] = network_profile_payload
            need_update = True
    if new_values.get("storage_data_disks") is not None:
        if (
            (existing_properties.get("storageProfile") is not None)
            and (existing_properties.get("storageProfile").get("dataDisks") is not None)
            and (
                compare_storage_data_disks_payload(
                    existing_properties.get("storageProfile").get("dataDisks"),
                    new_values.get("storage_data_disks"),
                )
            )
        ):
            updated_properties_payload["storageProfile"][
                "dataDisks"
            ] = convert_present_to_raw_data_disks(
                data_disks=merge_data_disks_payloads(
                    convert_raw_to_present_data_disks(
                        existing_properties.get("storageProfile").get("dataDisks")
                    ),
                    new_values.get("storage_data_disks"),
                )
            )
            need_update = True
    if new_values.get("storage_os_disk") is not None:
        if (
            (existing_properties.get("storageProfile") is not None)
            and (existing_properties.get("storageProfile").get("osDisk") is not None)
            and (
                compare_storage_os_disk_payload(
                    existing_properties.get("storageProfile").get("osDisk"),
                    new_values.get("storage_os_disk"),
                )
            )
        ):
            updated_properties_payload["storageProfile"][
                "osDisk"
            ] = convert_present_to_raw_os_disk(
                os_disk=merge_dictionary_payloads(
                    convert_raw_to_present_os_disk(
                        existing_properties.get("storageProfile").get("osDisk")
                    ),
                    new_values.get("storage_os_disk"),
                )
            )
            need_update = True
    if new_values.get("storage_image_reference") is not None:
        if (
            (existing_properties.get("storageProfile") is not None)
            and (
                existing_properties.get("storageProfile").get("imageReference")
                is not None
            )
            and (
                compare_storage_image_reference_payload(
                    existing_properties.get("storageProfile").get("imageReference"),
                    new_values.get("storage_image_reference"),
                )
            )
        ):
            updated_properties_payload["storageProfile"][
                "imageReference"
            ] = convert_present_to_raw_image_reference(
                storage_image_reference=merge_dictionary_payloads(
                    convert_raw_to_present_image_reference(
                        existing_properties.get("storageProfile").get("imageReference")
                    ),
                    new_values.get("storage_image_reference"),
                )
            )
            need_update = True
    if new_values.get("os_profile") is not None:
        if (existing_properties.get("osProfile") is not None) and (
            compare_os_profile_payload(
                existing_properties.get("osProfile"), new_values.get("os_profile")
            )
        ):
            updated_properties_payload["osProfile"] = convert_present_to_raw_os_profile(
                os_profile=merge_dictionary_payloads(
                    convert_raw_to_present_os_profile(
                        existing_properties.get("osProfile")
                    ),
                    new_values.get("os_profile"),
                )
            )
            need_update = True

    new_payload["properties"] = updated_properties_payload

    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values.get("tags")
        need_update = True

    if need_update:
        result["ret"] = new_payload
    return result


def convert_present_to_raw_network_interfaces(network_interface_ids: List[str]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        network_interface_ids(list(str)) : List of Network Interface Ids

    Returns:
        Network Interface Ids List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    network_interface_id_counter = 0
    network_interfaces_list: List = []
    for network_interface_id in network_interface_ids:
        network_interfaces_payload = {"id": network_interface_id}
        primary_network_payload = {}
        if network_interface_id_counter == 0:
            primary_network_payload["primary"] = True
        else:
            primary_network_payload["primary"] = False
        network_interfaces_payload["properties"] = primary_network_payload
        network_interfaces_list.append(network_interfaces_payload)
        network_interface_id_counter = network_interface_id_counter + 1
    return network_interfaces_list


def convert_present_to_raw_image_reference(storage_image_reference: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        storage_image_reference(Dict(str, Any)) : Specifies information about the image to use in VM creation/update

    Returns:
        Storage Image Reference Dict[str,any] in the format of an Azure PUT operation payload.
    """
    storage_image_reference_payload = {
        "publisher": storage_image_reference.get("image_publisher"),
        "offer": storage_image_reference.get("image_offer"),
        "sku": storage_image_reference.get("image_sku"),
        "version": storage_image_reference.get("image_version"),
    }
    return storage_image_reference_payload


def convert_present_to_raw_os_disk(os_disk: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        os_disk(Dict(str, Any)) : Specifies information about the operating system disk used by the virtual machine.

    Returns:
        OS Disk Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    os_disk_payload = {
        "name": os_disk.get("disk_name"),
        "diskSizeGB": os_disk.get("disk_size_in_GB"),
        "caching": os_disk.get("disk_caching"),
        "createOption": os_disk.get("disk_create_option"),
        "deleteOption": os_disk.get("disk_delete_option"),
        "managedDisk": {
            "id": os_disk.get("disk_id"),
            "storageAccountType": os_disk.get("storage_account_type"),
        },
    }
    return os_disk_payload


def convert_present_to_raw_data_disks(data_disks: List[Dict[str, Any]]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        data_disks(list(Dict[str, Any]) : List of Data Disk payload for VM

    Returns:
        List of Data Disk payload List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    data_disks_list: List = []
    for data_disk in data_disks:
        data_disks_payload = {
            "name": data_disk.get("disk_name"),
            "diskSizeGB": data_disk.get("disk_size_in_GB"),
            "lun": data_disk.get("disk_logical_unit_number"),
            "caching": data_disk.get("disk_caching"),
            "createOption": data_disk.get("disk_create_option"),
            "deleteOption": data_disk.get("disk_delete_option"),
            "managedDisk": {
                "id": data_disk.get("disk_id"),
                "storageAccountType": data_disk.get("storage_account_type"),
            },
        }
        data_disks_list.append(data_disks_payload)
    return data_disks_list


def convert_present_to_raw_os_profile(os_profile: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        os_profile(Dict[str, Any]) : Operating System profile for VM creation

    Returns:
        OS Profile payload : Dict[str,any] in the format of an Azure PUT operation payload.
    """
    os_profile_payload = {
        "adminUsername": os_profile.get("admin_username"),
        "adminPassword": os_profile.get("admin_password"),
        "computerName": os_profile.get("computer_name"),
    }
    return os_profile_payload


def convert_raw_to_present_os_profile(os_profile: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        os_profile(Dict, optional): OS Profile payload in a virtual machine resource

    Returns:
         OS Profile payload that contains the parameters that match respective present function's input format.
    """
    os_profile_payload = {
        "admin_username": os_profile.get("adminUsername"),
        "computer_name": os_profile.get("computerName"),
    }
    return os_profile_payload


def convert_raw_to_present_network_interface(network_interfaces: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        network_interfaces(list[Dict], optional): Resource List of Network Interfaces in a virtual machine resource.

    Returns:
         A Network Interface Id List that contains the parameters that match respective present function's input format.
    """
    present_network_interfaces: List = []
    for network_interface in network_interfaces:
        present_network_interfaces.append(network_interface.get("id"))
    return present_network_interfaces


def convert_raw_to_present_image_reference(image_reference: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        image_reference(Dict, optional): Image Reference payload in a virtual machine resource.

    Returns:
         Image Reference payload contains the parameters that match respective present function's input format.
    """
    image_reference_payload = {
        "image_sku": image_reference.get("sku"),
        "image_publisher": image_reference.get("publisher"),
        "image_version": image_reference.get("version"),
        "image_offer": image_reference.get("offer"),
    }
    return image_reference_payload


def convert_raw_to_present_os_disk(os_disk: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        os_disk(Dict, optional): OS Disk payload in a virtual machine resource.

    Returns:
         OS Disk payload that contains the parameters that match respective present function's input format.
    """
    os_disk_payload = {
        "disk_name": os_disk.get("name"),
        "disk_caching": os_disk.get("caching"),
    }
    if os_disk.get("diskSizeGB") is not None:
        os_disk_payload["disk_size_in_GB"] = os_disk.get("diskSizeGB")
    if (
        os_disk.get("managedDisk") is not None
        and os_disk.get("managedDisk").get("storageAccountType") is not None
    ):
        os_disk_payload["storage_account_type"] = os_disk.get("managedDisk").get(
            "storageAccountType"
        )
    if (
        os_disk.get("managedDisk") is not None
        and os_disk.get("managedDisk").get("id") is not None
    ):
        os_disk_payload["disk_id"] = os_disk.get("managedDisk").get("id")
    if os_disk.get("createOption") is not None:
        os_disk_payload["disk_create_option"] = os_disk.get("createOption")
    if os_disk.get("deleteOption") is not None:
        os_disk_payload["disk_delete_option"] = os_disk.get("deleteOption")
    return os_disk_payload


def convert_raw_to_present_data_disks(data_disks: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        data_disks(list[Dict], optional): Resource List of Data Disks in a virtual machine resource.

    Returns:
         A Data Disk List that contains the parameters that match respective present function's input format.
    """
    present_data_disks_payload: List = []
    for data_disk in data_disks:
        data_disk_payload = {
            "disk_name": data_disk.get("name"),
            "disk_logical_unit_number": data_disk.get("lun"),
            "disk_caching": data_disk.get("caching"),
        }
        if data_disk.get("diskSizeGB") is not None:
            data_disk_payload["disk_size_in_GB"] = data_disk.get("diskSizeGB")
        if (
            data_disk.get("managedDisk") is not None
            and data_disk.get("managedDisk").get("storageAccountType") is not None
        ):
            data_disk_payload["storage_account_type"] = data_disk.get(
                "managedDisk"
            ).get("storageAccountType")
        if (
            data_disk.get("managedDisk") is not None
            and data_disk.get("managedDisk").get("id") is not None
        ):
            data_disk_payload["disk_id"] = data_disk.get("managedDisk").get("id")
        if data_disk.get("createOption") is not None:
            data_disk_payload["disk_create_option"] = data_disk.get("createOption")
        if data_disk.get("deleteOption") is not None:
            data_disk_payload["disk_delete_option"] = data_disk.get("deleteOption")
        present_data_disks_payload.append(data_disk_payload)
    return present_data_disks_payload


def compare_network_interface_payload(
    existing_network_interface_list: List[Dict[str, Any]],
    network_interface_ids: List[str],
):
    """
    Compares network interface payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_network_interface_list(list[Dict]): Existing Network Interface payload
        network_interface_ids(list[str]): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    if len(network_interface_ids) != len(existing_network_interface_list):
        return True
    existing_network_interface_id_list = []
    for existing_network_interface in existing_network_interface_list:
        existing_network_interface_id_list.append(existing_network_interface.get("id"))
    return set(existing_network_interface_id_list) != set(network_interface_ids)


def compare_storage_os_disk_payload(
    existing_storage_os_disk_payload: Dict[str, Any],
    new_storage_os_disk: Dict[str, Any],
):
    """
    Compares OS Disk payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_storage_os_disk_payload[Dict]: Existing Storage OS Disk payload
        new_storage_os_disk[Dict]: Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    os_disk_present_converted_payload = convert_raw_to_present_os_disk(
        existing_storage_os_disk_payload
    )
    return compare_update_dict_payload(
        os_disk_present_converted_payload, new_storage_os_disk
    )


def compare_storage_data_disks_payload(
    existing_storage_data_disks_payload: List[Dict[str, Any]],
    new_storage_data_disks: List[Dict[str, Any]],
):
    """
    Compares Data Disks payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_storage_data_disks_payload(list[Dict]): Existing Storage Data Disk payload
        new_storage_data_disks(list[str]): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    data_disks_present_converted_payload = convert_raw_to_present_data_disks(
        existing_storage_data_disks_payload
    )
    if len(data_disks_present_converted_payload) != len(new_storage_data_disks):
        return True
    existing_payload_map = {}
    for element in data_disks_present_converted_payload:
        existing_payload_map[element.get("disk_name")] = element

    return compare_update_dict_list_payload(
        existing_payload_map, new_storage_data_disks
    )


def compare_storage_image_reference_payload(
    existing_storage_image_reference_payload: Dict[str, Any],
    new_storage_image_reference: Dict[str, Any],
):
    """
    Compares Storage Image Reference payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_storage_image_reference_payload(Dict): Existing Storage Image Reference payload
        new_storage_image_reference(Dict): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    storage_image_reference_present_converted_payload = (
        convert_raw_to_present_image_reference(existing_storage_image_reference_payload)
    )
    return compare_update_dict_payload(
        storage_image_reference_present_converted_payload, new_storage_image_reference
    )


def compare_os_profile_payload(
    existing_os_profile_payload: Dict[str, Any], new_os_profile: Dict[str, Any]
):
    """
    Compares OS Profile payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_os_profile_payload(Dict): Existing OS Profile payload
        new_os_profile(Dict): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    os_profile_present_converted_payload = convert_raw_to_present_os_profile(
        existing_os_profile_payload
    )
    new_os_profile_copy = copy.deepcopy(new_os_profile)
    # azure payload does not contain admin_password
    new_os_profile_copy.pop("admin_password", None)
    return compare_update_dict_payload(
        os_profile_present_converted_payload, new_os_profile_copy
    )


def compare_update_dict_payload(
    existing_payload: Dict[str, Any], update_payload: Dict[str, Any]
):
    """
    Compares the payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_payload(Dict): Existing payload
        update_payload(Dict): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    for parameter in update_payload:
        if parameter in existing_payload:
            if update_payload.get(parameter) != existing_payload.get(parameter):
                return True
        else:
            return True
    return False


def compare_update_dict_list_payload(
    existing_payload_map: Dict[str, Any], update_payload: List[Dict[str, Any]]
):
    """
    Compares the payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_payload_map(Dict): Existing payload Map
        update_payload(List(Dict)): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    for update_element in update_payload:
        disk_name = update_element.get("disk_name")
        if disk_name in existing_payload_map:
            if compare_update_dict_payload(
                existing_payload_map.get(disk_name), update_element
            ):
                return True
        else:
            return True
    return False


def merge_dictionary_payloads(
    existing_payload: Dict[str, Any], update_payload: Dict[str, Any]
):
    """
    Merge the input payload with the payload from Azure Get Response
    Returns the merged payload

    Args:
        existing_payload(Dict): Existing payload From Azure Get API
        update_payload(Dict): Present value which will be given as input

    Returns:
        A Dict of the merged Payload
    """
    existing_payload_copy = copy.deepcopy(existing_payload)
    for update_element_key in update_payload:
        if update_payload.get(update_element_key) is not None:
            existing_payload_copy[update_element_key] = update_payload.get(
                update_element_key
            )
    return existing_payload_copy


def merge_data_disks_payloads(
    existing_payload: List[Dict[str, Any]], update_payload: List[Dict[str, Any]]
):
    """
    Merge the input payload with the payload from Azure Get Response for Data Disks
    Returns the merged Data Disk payload

    Args:
        existing_payload(Dict): Existing payload Map
        update_payload(Dict): Present value which will be given as input

    Returns:
        A Dict of the merged Data Disks Payload
    """
    existing_payload_map = {}
    existing_merged_payload = []
    for element in existing_payload:
        existing_payload_map[element.get("disk_name")] = element
    for update_element in update_payload:
        disk_name = update_element.get("disk_name")
        if disk_name in existing_payload_map:
            # Merge Data
            existing_merged_payload.append(
                merge_dictionary_payloads(
                    existing_payload_map.get(disk_name), update_element
                )
            )
        else:
            # Keep new input data
            existing_merged_payload.append(update_element)
    return existing_merged_payload



"""State module for managing Attach Subscription."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    management_group_id: str,
    subscription_id: str,
    resource_id: str = None,
) -> Dict:
    r"""Create or update attached subscription to management group.

    Args:
        name(str): The identifier for this state.
        management_group_id(str): management group id to which subscription needs to be associated.
        subscription_id(str): subscription unique id
        resource_id(str, Optional): resource unique id

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            attach_subscription_to_mg_present:
              azure.subscription.attach_subscriptions.present:
                - name: value
                - management_group_id: value
                - subscription_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if resource_id is None:
        resource_id = f"/providers/Microsoft.Management/managementGroups/{management_group_id}/subscriptions/{subscription_id}"
    response_get = await hub.exec.azure.subscription.attach_subscriptions.get(
        ctx,
        resource_id=resource_id,
        raw=True,
    )
    if response_get["result"]:
        if response_get["ret"] is None:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "management_group_id": management_group_id,
                        "subscription_id": subscription_id,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.subscription.attach_subscriptions '{name}'"
                )
                return result
            else:
                # PUT operation to attach subscription to management group
                payload = {}
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not attach subscription to given management group : {response_put['comment']}"
                        f" {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

            result[
                "new_state"
            ] = hub.tool.azure.subscription.subscriptions.convert_raw_attach_subscription_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                management_group_id=management_group_id,
                subscription_id=subscription_id,
                resource_id=resource_id,
            )
            result["comment"].append(
                f"Created azure.subscription.attach_subscriptions '{name}'"
            )
            return result
        else:
            result[
                "old_state"
            ] = hub.tool.azure.subscription.subscriptions.convert_raw_attach_subscription_to_present(
                resource=response_get["ret"],
                idem_resource_name=name,
                management_group_id=management_group_id,
                subscription_id=subscription_id,
                resource_id=resource_id,
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"No update required azure.subscription.attach_subscriptions '{name}'"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.subscription.attach_subscriptions {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def absent(
    hub,
    ctx,
    name: str,
    management_group_id: str,
    subscription_id: str,
) -> Dict:
    r"""De-Associate subscription from management group.

    Args:
        name(str): The identifier for this state.
        management_group_id(str): management group unique id.
        subscription_id(str): subscription unique id.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            detach_subscription_from_mg_absent:
              azure.subscription.attach_subscriptions.absent:
                - name: value
                - management_group_id: value
                - subscription_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    resource_id = f"/providers/Microsoft.Management/managementGroups/{management_group_id}/subscriptions/{subscription_id}"
    response_get = await hub.exec.azure.subscription.attach_subscriptions.get(
        ctx,
        resource_id=resource_id,
        raw=True,
    )
    if response_get["result"]:
        if response_get["ret"]:
            result[
                "old_state"
            ] = hub.tool.azure.subscription.subscriptions.convert_raw_attach_subscription_to_present(
                resource=response_get["ret"],
                idem_resource_name=name,
                management_group_id=management_group_id,
                subscription_id=subscription_id,
                resource_id=resource_id,
            )
            if ctx.get("test", False):
                result["comment"].append(
                    f"Would de-associates azure.subscription.attach_subscriptions '{name}'"
                )
                return result

            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
                success_codes=[200, 202],
            )

            if not response_delete["result"]:
                result["result"] = False
                result["comment"] = [response_delete["comment"], response_delete["ret"]]
                return result

            result["comment"].append(
                f"De-associated azure.subscription.attach_subscriptions '{name}'"
            )
            return result
        else:
            # If Azure returns 'Not Found' error, it means the management group is not associated with given subscription.
            result["comment"].append(
                f"azure.subscription.attach_subscriptions '{name}' is not associated with management group"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get subscription attached to management group : '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all subscriptions with respective alias details.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            $ idem describe azure.subscription.attach_subscriptions
    """
    result = {}
    ret_list = await hub.exec.azure.subscription.attach_subscriptions.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe attach subscriptions {ret['comment']}")
        return {}
    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.subscription.attach_subscriptions.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result
